<?php
/**
 * Mondial Relay Official Shipping method
 *
 * @package mondialrelay
 */
declare(strict_types=1);

defined( 'ABSPATH' ) || exit;

/**
 * Mondial Relay Official Shipping method
 */
class MRWP_Shipping_Method extends \WC_Shipping_Method
{
	/**
	 * ID of shipping method
	 */
	public static $method_id = 'mondialrelay_official_shipping';

	/**
	 * Constructor
	 *
	 * @param integer $instance_id  WooCommerce Shipping instance
	 */
	public function __construct( $instance_id = 0 )
	{
		$this->id                 = self::$method_id;
		$this->instance_id        = absint( $instance_id );

		// Init before using ->get_instance_option()
		$this->init_form_fields();

		// Title column
		$this->title              = $this->get_instance_option( 'title', __( 'Mondial Relay Officiel', 'mondialrelay-wordpress' ) );

		// Description column
		$this->method_title       = __( 'Mondial Relay Officiel', 'mondialrelay-wordpress' );
		$this->method_description = __( 'Mondial Relay - Extension WordPress officielle', 'mondialrelay-wordpress' );

		$this->supports = array(
			'shipping-zones',
			'instance-settings',
			'instance-settings-modal',
		);

		// Init shipping mode if this is an uninitialized method
		if ( 0 != $instance_id ) {
			$this->maybe_init_mode();
		}

		// Save settings in admin
		add_action( 'woocommerce_update_options_shipping_' . $this->id, array( $this, 'process_admin_options' ) );
	}

	/**
	 * Filter: add our shipping method to the list
	 *
	 * @param array $methods  WooCommerce shipping methods list
     *        array index = id from corresponding Shipping_Method class
	 */
	public static function add_shipping_method( array $methods )
	{
        $method = array( self::$method_id => 'MRWP_Shipping_Method' );

		return $method + $methods;
	}

	/**
	 * Init fields
	 *
	 * Add fields to the shipping settings page.
	 */
	public function init_form_fields()
	{
		if ( 'yes' === get_option( 'woocommerce_prices_include_tax' ) ) {
			$memo_tax  = __( 'Note&nbsp;: les tarifs de la boutique sont saisis TTC.', 'mondialrelay-wordpress' );
		} else {
			$memo_tax  = __( 'Note&nbsp;: les tarifs de la boutique sont saisis HT.', 'mondialrelay-wordpress' );
		}
		$free_shipping_desc = '<span class="tax">' . $memo_tax . '</span>';

		$url   = plugin_dir_url( __DIR__ ) . 'public/img/mondial-relay-logo.png';
		$style = 'margin-left: 16px; max-width: 32px; display:inline; vertical-align: middle;';
		$logo  = '<img src="' . $url . '" style="' . $style . '"> ';

		$this->instance_form_fields = array(
			'title' => array(
				'title'       => __( 'Nom', 'mondialrelay-wordpress' ),
				'description' => __( 'Nom de la méthode d’expédition à afficher sur le site', 'mondialrelay-wordpress' ),
				'type'        => 'text',
				'default'     => __( 'Mondial Relay', 'mondialrelay-wordpress' ),
				'desc_tip'    => true,
			),
			'display_logo' => array(
				'title'       => __( 'Afficher le logo', 'mondialrelay-wordpress' ),
				'type'        => 'checkbox',
				'label'       => __( 'Cocher pour afficher le logo Mondial Relay', 'mondialrelay-wordpress' ) . $logo,
				'default'     => 'yes'
			),
			'shipping_description' => array(
				'title'       => __( 'Description additionnelle', 'mondialrelay-wordpress' ),
				'description' => __( 'Ce texte optionnel s’affiche sous le nom de la méthode d’expédition. Ex. « livraison sous 3 jours ouvrés »', 'mondialrelay-wordpress' ),
				'type'        => 'text',
				'default'     => '',
				'desc_tip'    => true,
			),
			'free_shipping_from' => array(
				'title'       => __( 'Offrir la livraison à partir de ce montant (optionnel)', 'mondialrelay-wordpress' ),
				'description' => $free_shipping_desc,
				'type'        => 'text',
				'class'       => 'price',
				'default'     => '',
			),
			'title_free' => array(
				'title'       => __( 'Nom si la livraison est offerte', 'mondialrelay-wordpress' ),
				'description' => __( 'Nom de la méthode d’expédition si la livraison est gratuite', 'mondialrelay-wordpress' ),
				'type'        => 'text',
				'default'     => '',
				'desc_tip'    => true,
			),
			'method' => array(
				'title'       => __( 'Mode de livraison Mondial Relay', 'mondialrelay-wordpress' ),
				'type'        => 'select',
				'default'     => '24R',
				'options'     => MRWP_modes::all_labels_with_hmr(),
			),
			'conditions' => array(
				'type'              => 'rates_table', // Calls $this->generate_rates_table_html()
				'sanitize_callback' => array( $this, 'sanitize_rates' ),
			),
		);
	}

	/**
	 * Generate the rates table
	 *
	 * @return string HTML table
	 */
	public function generate_rates_table_html()
	{
		// Retrieve shipping rates
		$conditions = $this->get_instance_option( 'conditions' );

		// Set default initial values if table is empty
		if ( ! is_array( $conditions ) or count( $conditions ) < 1 ) {

			// $sample_values is used in the view
			$sample_values = true;

			$conditions = array();
			$conditions[0]['weight'] = '0.500';
			$conditions[0]['tare']   = '';
			$conditions[0]['price']  = '';

			$conditions[1]['weight'] = '1.000';
			$conditions[1]['tare']   = '';
			$conditions[1]['price']  = '';

			$conditions[2]['weight'] = '2.000';
			$conditions[2]['tare']   = '';
			$conditions[2]['price']  = '';

			$conditions[3]['weight'] = '3.000';
			$conditions[3]['tare']   = '';
			$conditions[3]['price']  = '';

			$conditions[4]['weight'] = '4.000';
			$conditions[4]['tare']   = '';
			$conditions[4]['price']  = '';

			$conditions[5]['weight'] = '5.000';
			$conditions[5]['tare']   = '';
			$conditions[5]['price']  = '';

			$conditions[6]['weight'] = '7.000';
			$conditions[6]['tare']   = '';
			$conditions[6]['price']  = '';

			$conditions[7]['weight'] = '10.000';
			$conditions[7]['tare']   = '';
			$conditions[7]['price']  = '';

			$conditions[8]['weight'] = '15.000';
			$conditions[8]['tare']   = '';
			$conditions[8]['price']  = '';

			$conditions[9]['weight'] = '20.000';
			$conditions[9]['tare']   = '';
			$conditions[9]['price']  = '';

			$conditions[10]['weight'] = '30.000';
			$conditions[10]['tare']   = '';
			$conditions[10]['price']  = '';
		} else {
			$sample_values = false;
		}

		// Input fields name prefix
		// Note: the final ] is mandatory for proper submission of the form
		//       otherwise we get a weirdly indexed array in $_POST
		$prefix = 'woocommerce_' . $this->id . '_conditions]';

		// Start buffering
		ob_start();

		require wp_normalize_path( __DIR__ . '/rates-table.inc' );

		return ob_get_clean();
	}

	/**
	 *  Display currency before or after input field
	 */
	private function display_currency( string $where )
	{
		// Retrieve currency position eg 'right' ou 'right_space'
		$currency_pos = get_option( 'woocommerce_currency_pos' );

		if ( strpos( $currency_pos, 'right' ) === 0 ) {
			if ( 'after' == $where ) {
				echo '&nbsp;' . html_entity_decode( get_woocommerce_currency_symbol() );
			}
		} else {
			if ( 'before' == $where ) {
				echo html_entity_decode( get_woocommerce_currency_symbol() ) . '&nbsp;';
			}
		}
	}

	/**
	 * Sanitize values from the form
	 *
	 * Example returned values
	 *
	 * [0] => Array
	 *     [weight] => 0,500
	 *     [tare]   => 0.050
	 *     [price]  => 3.99
	 *
	 * @param  array $table  Form data
	 */
	public function sanitize_rates( $table )
	{
		$conditions = array();
		if ( is_array( $table ) ) {
			foreach ( $table as $row ) {

				// Ignore empty lines
				if ( '' == trim( $row['weight'] ) ) {
					continue;
				}

				$condition = array();
				$condition['weight'] = $this->parse_float( $row['weight'] );
				$condition['tare']   = $this->parse_float( $row['tare'] );
				$condition['price']  = $this->parse_float( $row['price'], false );
				$conditions[] = $condition;
			}
		}

		return $conditions;
	}

	/**
	 * Processes and saves global shipping method options in the admin area.
	 *
	 * @return bool was anything saved?
	 */
	public function process_admin_options()
	{
		// Check we are processing the correct form for this instance.
		if ( ! isset( $_REQUEST['instance_id'] ) || absint( $_REQUEST['instance_id'] ) !== $this->instance_id ) {
			return false;
		}

		$this->init_instance_settings();

		$post_data = $this->get_post_data();
		foreach ( $this->get_instance_form_fields() as $key => $field ) {

			// Ignore title fields
			if ( 'title' !== $this->get_field_type( $field ) ) {
				try {
					$value = $this->get_field_value( $key, $field, $post_data );
					$this->instance_settings[ $key ] = $value;

					// Update the legacy option update_ids_livraison
					if ( 'method' == $key ) {
						$this->update_ids_livraison( $value );
					}
				} catch ( Exception $e ) {
					$this->add_error( $e->getMessage() );
				}
			}
		}

		return update_option(
			$this->get_instance_option_key(),
			apply_filters( 'woocommerce_shipping_' . $this->id . '_instance_settings_values', $this->instance_settings, $this ),
			'yes'
		);
	}

	private function update_ids_livraison( string $value )
	{
		$ids    = (array)get_option( 'mondialrelay_ids_livraison' );
		$our_id = $this->id . ':' . $this->instance_id;

		$key_exists = false;
		foreach ( $ids as $key => $id ) {
			if ( 0 === strpos( $id, $our_id ) ) {
				$ids[ $key ] = $our_id . '-' . $value;
				$key_exists = true;
				break;
			}
		}

		// If we didn’t find our id, it’s a new zone ; add it.
		if ( ! $key_exists ) {
			$ids[] = $our_id . '-' . $value;
		}
		update_option( 'mondialrelay_ids_livraison', $ids );
	}

	private function maybe_init_mode()
	{
		$ids    = (array)get_option( 'mondialrelay_ids_livraison' );
		$our_id = $this->id . ':' . $this->instance_id;

		foreach ( $ids as $key => $id ) {
			if ( 0 === strpos( $id, $our_id ) ) {

				// This instance already has a delivery mode defined
				return;
			}
		}

		// If we didn’t find our id, it’s a new zone ; add it.
		if ( ! $key_exists ) {
			$ids[] = $our_id . '-' . '24R';
		}
		update_option( 'mondialrelay_ids_livraison', $ids );
	}

	/**
	 * Calculate shipping price
	 *
	 * @param  array $package Package
	 *
	 * Note: don’t declare parameter type (array) as it would cause a compatibility warning with
	 *       the declaration in class WC_Shipping_Method
	 */
	public function calculate_shipping( $package = array() )
	{
		$logger = wc_get_logger();

		// Compute total weight in kg (whatever the shop unit is)
		$net_weight = 0.000;
		foreach ( $package['contents'] as $item_id => $values ) {
			$product          = $values['data'];
			$one_item         = wc_get_weight( $product->get_weight(), 'kg' );
			$net_weight = $net_weight + $one_item * $values['quantity'];
		}

		// Is a free shipping coupon applied ?
		$coupons = $package['applied_coupons'];
		foreach ( $coupons as $code ) {
			$coupon = new WC_Coupon( $code );
			if ( $coupon->get_free_shipping() ) {
				$this->add_shipping_rate( 0.00, $package );

				return;
			}
		}

		// Get cart amount
		$this->cart_subtotal = (float)$package['cart_subtotal'];

		// free_shipping_from can be:
		//   empty => shipping is never free
		//   0.00  => shipping is always free
		//  99.00  => shipping is free if cart value is more than 99 (for example)
		$free_shipping_from = $this->get_instance_option( 'free_shipping_from' );

		// Check for free delivery
		if ( '' !== $free_shipping_from ) {
			$minimum = $this->parse_float( $free_shipping_from );
			if ( $this->cart_subtotal >= $minimum ) {
				$this->add_shipping_rate( 0.00, $package );

				return;
			}
		}

		// Delivery is not free, retrieve user pricing table
		$conditions = $this->get_instance_option( 'conditions' );

		// Check we have at least 1 line
		if ( is_array( $conditions ) and count( $conditions ) > 0 ) {

			// Test each table line one by one, select minimum price
			foreach ( $conditions as $condition ) {
				if ( $net_weight + (float)$condition['tare'] <= $condition['weight'] ) {
					$this->add_shipping_rate( $condition['price'], $package );

					return;
				}
			}

			// No rate found, bail out
			return;
		}
	}

	/**
	 * Returns tare in kg
	 *
	 * @param  int $contents_weight Total weight of products in grams
	 *
	 * Note: weight unit is kg in this method
	 */
	public function get_tare_for_weight( int $net_weight )
	{
		$conditions = $this->get_instance_option( 'conditions' );

		// Check we have at least 1 line
		if ( is_array( $conditions ) and count( $conditions ) > 1 ) {

			// Test each table line one by one, select tare
			foreach ( $conditions as $condition ) {
				if ( $net_weight <= $condition['weight'] ) {

					return (float)$condition['tare'];
				}
			}
		}

		return 0.000;
	}

	/**
	 * Register our shipping price with WooCommerce
	 *
	 * @param  float $price Shipping price
	 */
	protected function add_shipping_rate( $price, $package )
	{
		$title = $this->get_option( 'title' );

		// If shipping is free and option is set, use « title_free » instead
		$title_free = trim( $this->get_option( 'title_free' ) );
		if ( 0.00 == $price and '' != $title_free ) {
			$title = $title_free;
		}

		$rate = array(
			'id'      => $this->get_rate_id(),
			'label'   => $title,
			'cost'    => $price,
			'package' => $package,
		);

		// Register the rate
		$this->add_rate( $rate );
	}

	/**
	 * Optionally add Mondial Relay logo and description to shipping label in cart
	 *
	 * @param  string $label Original label
	 */
	public static function embellish_label( string $label, $rate )
	{
        $logo = '';
        $desc = '';
		if ( self::$method_id == $rate->method_id ) {
			$method = new MRWP_Shipping_Method( $rate->instance_id );
			if ( 'yes' == $method->get_option( 'display_logo' ) ) {
				$url   = plugin_dir_url( __DIR__ ) . 'public/img/mondial-relay-logo.png';
				$style = 'max-width: 32px; display:inline; vertical-align: middle;';
				$logo  = '<img src="' . $url . '" style="' . $style . '" class="shipping_logo"> ';
			}
			$option = trim( $method->get_option( 'shipping_description' ) );
			if ( '' != $option ) {
				$style = 'font-style: italic;';
				$desc = '<br><span style="' . $style . '" class="shipping_description">' . esc_html( $option ) . '</span>';
			}
        }

        return $logo . $label . $desc;
    }

	/**
	 * Parses user-entered string as a floating point value
	 * Empty value is allowed (coded as empty string '')
	 *
	 * @param  string $value     Value from form
	 * @param  bool   $empty_ok  Allow empty value
	 *
	 * @return mixed             Sanitized value, valid float or empty string
	 */
	protected function parse_float( $value, $empty_ok = true )
	{
		$value = trim( $value );

		// Empty?
		if ( $empty_ok and '' == $value ) {
			return '';
		}

		// Maybe replace decimal comma with decimal point
		$value = str_replace( ',', '.', $value );
		$value = floatval( $value );

		return $value;
	}

	/**
	 * Formats weight for use in rates table
	 * Empty value is allowed (coded as empty string '')
	 *
	 * @param  mixed $weight
	 *
	 * @return string
	 */
	protected function format_weight( $weight )
	{
		if ( is_string( $weight ) and '' == trim( $weight ) ) {
			return '';
		}

		$value = number_format( (float)$weight, 3, wc_get_price_decimal_separator(), wc_get_price_thousand_separator() );

		return $value;
	}

}
